"""A creation workflow for creating a new interconnect between a switch and a router."""

from typing import Annotated
from uuid import uuid4

from annotated_types import Len
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic import AfterValidator, ConfigDict
from pydantic_forms.validators import Divider, ReadOnlyField

from gso.annotations.interfaces import (
    JuniperAEInterface,
    JuniperLAGMember,
    JuniperPhyInterface,
    LAGMember,
    PhysicalPortCapacity,
    validate_interface_names_are_unique,
)
from gso.annotations.tt_number import TTNumber
from gso.products.product_blocks.lan_switch_interconnect import (
    LanSwitchInterconnectAddressSpace,
    LanSwitchInterconnectInterfaceBlockInactive,
)
from gso.products.product_types.lan_switch_interconnect import LanSwitchInterconnectInactive
from gso.products.product_types.router import Router
from gso.products.product_types.switch import Switch
from gso.services.partners import get_partner_by_name
from gso.utils.helpers import (
    active_router_selector,
    active_switch_selector,
    available_interfaces_choices,
    available_lags_choices,
)
from gso.utils.shared_enums import Vendor
from gso.utils.types.interfaces import (
    JuniperAEInterface,
    JuniperLAGMember,
    JuniperPhyInterface,
    LAGMember,
    PhysicalPortCapacity,
    validate_interface_names_are_unique,
)
from gso.utils.types.tt_number import TTNumber


def _initial_input_form(product_name: str) -> FormGenerator:
    class CreateLANSwitchInterconnectForm(FormPage):
        model_config = ConfigDict(title=product_name)

        tt_number: TTNumber
        router_side: active_router_selector()  # type: ignore[valid-type]
        switch_side: active_switch_selector()  # type: ignore[valid-type]
        address_space: LanSwitchInterconnectAddressSpace
        description: str
        minimum_link_count: int
        divider: Divider
        vlan_id: ReadOnlyField(111, default_type=int)  # type: ignore[valid-type]

    user_input = yield CreateLANSwitchInterconnectForm
    router = Router.from_subscription(user_input.router_side)

    if router.router.vendor == Vendor.NOKIA:

        class NokiaLAGMemberA(LAGMember):
            interface_name: available_interfaces_choices(  # type: ignore[valid-type]
                router.subscription_id,
                PhysicalPortCapacity.TEN_GIGABIT_PER_SECOND,
            )

        router_side_ae_member_list = Annotated[
            list[NokiaLAGMemberA],
            AfterValidator(validate_interface_names_are_unique),
            Len(min_length=user_input.minimum_link_count),
        ]
    else:
        router_side_ae_member_list = Annotated[  # type: ignore[assignment, misc]
            list[JuniperLAGMember],
            AfterValidator(validate_interface_names_are_unique),
            Len(min_length=user_input.minimum_link_count),
        ]

    class InterconnectRouterSideForm(FormPage):
        model_config = ConfigDict(title="Please enter interface names and descriptions for the router side.")

        router_side_iface: available_lags_choices(user_input.router_side) or JuniperAEInterface  # type: ignore[valid-type]
        router_side_ae_members: router_side_ae_member_list

    router_side_input = yield InterconnectRouterSideForm

    switch_side_ae_member_list = Annotated[
        list[JuniperLAGMember],
        AfterValidator(validate_interface_names_are_unique),
        Len(
            min_length=len(router_side_input.router_side_ae_members),
            max_length=len(router_side_input.router_side_ae_members),
        ),
    ]

    class InterconnectSwitchSideForm(FormPage):
        model_config = ConfigDict(title="Please enter interface names and descriptions for the switch side.")

        switch_side_iface: JuniperPhyInterface
        switch_side_ae_members: switch_side_ae_member_list

    switch_side_input = yield InterconnectSwitchSideForm

    return user_input.model_dump() | router_side_input.model_dump() | switch_side_input.model_dump()


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: str) -> State:
    """Create a new subscription object in the database."""
    subscription = LanSwitchInterconnectInactive.from_product_id(product, get_partner_by_name(partner)["partner_id"])

    return {"subscription": subscription}


@step("Initialize subscription")
def initialize_subscription(
    subscription: LanSwitchInterconnectInactive,
    description: str,
    address_space: LanSwitchInterconnectAddressSpace,
    minimum_link_count: int,
    router_side: UUIDstr,
    router_side_iface: JuniperPhyInterface,
    router_side_ae_members: list[dict],
    switch_side: UUIDstr,
    switch_side_iface: JuniperPhyInterface,
    switch_side_ae_members: list[dict],
) -> State:
    """Update the product model with all input from the operator."""
    subscription.lan_switch_interconnect.lan_switch_interconnect_description = description
    subscription.lan_switch_interconnect.address_space = address_space
    subscription.lan_switch_interconnect.minimum_links = minimum_link_count
    subscription.lan_switch_interconnect.router_side.node = Router.from_subscription(router_side).router
    subscription.lan_switch_interconnect.router_side.ae_iface = router_side_iface
    for member in router_side_ae_members:
        subscription.lan_switch_interconnect.router_side.ae_members.append(
            LanSwitchInterconnectInterfaceBlockInactive.new(subscription_id=uuid4(), **member)
        )
    subscription.lan_switch_interconnect.switch_side.node = Switch.from_subscription(switch_side).switch
    subscription.lan_switch_interconnect.switch_side.ae_iface = switch_side_iface
    for member in switch_side_ae_members:
        subscription.lan_switch_interconnect.switch_side.ae_members.append(
            LanSwitchInterconnectInterfaceBlockInactive.new(subscription_id=uuid4(), **member)
        )

    return {"subscription": subscription}


@workflow(
    "Create LAN switch interconnect",
    initial_input_form=wrap_create_initial_input_form(_initial_input_form),
    target=Target.CREATE,
)
def create_lan_switch_interconnect() -> StepList:
    """Create a new LAN interconnect between a Switch and a Router."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
