"""Promote a P router to a PE router."""

import json
from typing import Any

from orchestrator.config.assignee import Assignee
from orchestrator.forms import FormPage
from orchestrator.forms.validators import Label
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, UUIDstr
from orchestrator.utils.errors import ProcessFailureError
from orchestrator.utils.json import json_dumps
from orchestrator.workflow import StepList, begin, done, inputstep, step, workflow
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import ConfigDict, model_validator

from gso.annotations.tt_number import TTNumber
from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import Router
from gso.services.kentik_client import KentikClient, NewKentikDevice
from gso.services.lso_client import LSOState, lso_interaction
from gso.services.subscriptions import get_all_active_sites
from gso.utils.helpers import generate_inventory_for_active_routers
from gso.utils.shared_enums import Vendor
from gso.utils.types.tt_number import TTNumber
from gso.utils.workflow_steps import (
    add_all_p_to_pe_dry,
    add_all_p_to_pe_real,
    add_pe_mesh_to_pe_dry,
    add_pe_mesh_to_pe_real,
    add_pe_to_all_p_dry,
    add_pe_to_all_p_real,
    add_pe_to_pe_mesh_dry,
    add_pe_to_pe_mesh_real,
    check_l3_services,
    check_pe_ibgp,
    update_sdp_mesh_dry,
    update_sdp_mesh_real,
)


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Promote P router to PE router."""
    subscription = Router.from_subscription(subscription_id)

    class PromotePToPEForm(FormPage):
        model_config = ConfigDict(title=f"Promote {subscription.router.router_fqdn} to PE router?")

        tt_number: TTNumber

        @model_validator(mode="before")
        def router_must_be_nokia_p(cls, data: Any) -> Any:
            if not (subscription.router.router_role == RouterRole.P and subscription.router.vendor == Vendor.NOKIA):
                msg = "PE promotion workflow can only be run on Nokia P routers."
                raise ValueError(msg)

            return data

    user_input = yield PromotePToPEForm

    return user_input.model_dump() | {"subscription": subscription}


@step("Evacuate the router by setting isis_overload")
def set_isis_overload(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Evacuate the router by setting isis overload."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Set ISIS overload",
        "verb": "set_isis_overload",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("[DRY RUN] Deploy PE base config")
def deploy_pe_base_config_dry(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a dry run of adding the base config to the router."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - deploy PE base config",
        "verb": "deploy_pe_base_config",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
        "geant_sites": json.loads(json_dumps(get_all_active_sites())),
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Deploy PE base config")
def deploy_pe_base_config_real(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a real run of adding the base config to the router."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - deploy PE base config",
        "verb": "deploy_pe_base_config",
        "pe_router_list": generate_inventory_for_active_routers(RouterRole.PE)["all"]["hosts"],
        "geant_sites": json.loads(json_dumps(get_all_active_sites())),
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@inputstep("Prompt EARL insertion", assignee=Assignee.SYSTEM)
def prompt_insert_in_earl(subscription: dict[str, Any]) -> FormGenerator:
    """Wait for confirmation from an operator that the router has been inserted in EARL."""

    class EARLPrompt(FormPage):
        model_config = ConfigDict(title="Update RADIUS clients")

        info_label: Label = f"Please add the router {subscription["router"]["router_fqdn"]} to EARL."

    yield EARLPrompt

    return {}


@step("Create Kentik device")
def create_kentik_device(subscription: Router) -> State:
    """Create a new device in Kentik."""
    if not (
        subscription.router.router_site
        and subscription.router.router_site.site_name
        and subscription.router.router_site.site_tier
        and subscription.router.router_fqdn
    ):
        msg = "Router object is missing required properties."
        raise ProcessFailureError(msg)

    kentik_client = KentikClient()
    kentik_site = kentik_client.get_site_by_name(subscription.router.router_site.site_name)

    if not kentik_site:
        msg = "Site could not be found in Kentik."
        raise ProcessFailureError(msg, details=subscription.router.router_site.site_name)

    new_device = NewKentikDevice(
        device_name=subscription.router.router_fqdn,
        device_description=str(subscription.subscription_id),
        sending_ips=[str(subscription.router.router_lo_ipv4_address)],
        site_id=kentik_site["id"],
        device_snmp_ip=str(subscription.router.router_lo_ipv4_address),
        device_bgp_flowspec=False,
        device_bgp_neighbor_ip=str(subscription.router.router_lo_ipv4_address),
        device_bgp_neighbor_ip6=str(subscription.router.router_lo_ipv6_address),
    )
    kentik_device = kentik_client.create_device(new_device)

    return {"kentik_device": kentik_device}


@step("[DRY RUN] Remove P from all PEs")
def remove_p_from_pe_dry(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a dry run of removing the P router from all the PE routers."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Remove P-only router neighbour {subscription["router"]["router_fqdn"]} from PE routers P-only group",
        "verb": "remove_p_from_pe",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/update_ibgp_mesh.yaml",
        "inventory": generate_inventory_for_active_routers(RouterRole.PE),
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Remove P from all PEs")
def remove_p_from_pe_real(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Remove the P router from all the PE routers."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Remove P-only router neighbour {subscription["router"]["router_fqdn"]} from PE routers P-only group",
        "verb": "remove_p_from_pe",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/update_ibgp_mesh.yaml",
        "inventory": generate_inventory_for_active_routers(RouterRole.PE),
        "extra_vars": extra_vars,
    }


@step("[DRY RUN] Deploy routing instances")
def deploy_routing_instances_dry(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a dry run of deploying routing instances."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy PE base config",
        "verb": "deploy_routing_instances",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Deploy routing instances")
def deploy_routing_instances_real(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a real run of deploying routing instances."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Deploy PE base config",
        "verb": "deploy_routing_instances",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("Remove ISIS overload")
def remove_isis_overload(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Remove ISIS overload."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - Remove ISIS overload",
        "verb": "remove_isis_overload",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("Set router role to PE (Update subscription model)")
def update_subscription_model(subscription: Router) -> State:
    """Update the subscription model to set router role to PE."""
    subscription.router.router_role = RouterRole.PE

    return {"subscription": subscription}


@step("[DRY RUN] Delete default routes")
def delete_default_routes_dry(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a dry run of deleting the default routes."""
    extra_vars = {
        "dry_run": True,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Delete static default routes (part of P base-config)",
        "verb": "delete_default_routes",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@step("[FOR REAL] Delete default routes")
def delete_default_routes_real(subscription: dict[str, Any], tt_number: str, process_id: UUIDstr) -> LSOState:
    """Perform a real run of deleting the default routes."""
    extra_vars = {
        "dry_run": False,
        "subscription": subscription,
        "commit_comment": f"GSO_PROCESS_ID: {process_id} - TT_NUMBER: {tt_number} - "
        f"Delete static default routes (part of P base-config)",
        "verb": "delete_default_routes",
    }

    return {
        "playbook_name": "gap_ansible/playbooks/promote_p_to_pe.yaml",
        "inventory": {"all": {"hosts": {subscription["router"]["router_fqdn"]: None}}},
        "extra_vars": extra_vars,
    }


@workflow(
    "Promote P router to PE router",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def promote_p_to_pe() -> StepList:
    """Promote a P router to a PE router."""
    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> lso_interaction(set_isis_overload)
        >> lso_interaction(deploy_pe_base_config_dry)
        >> lso_interaction(deploy_pe_base_config_real)
        >> prompt_insert_in_earl
        >> create_kentik_device
        >> lso_interaction(remove_p_from_pe_dry)
        >> lso_interaction(remove_p_from_pe_real)
        >> lso_interaction(add_pe_mesh_to_pe_dry)
        >> lso_interaction(add_pe_mesh_to_pe_real)
        >> lso_interaction(add_pe_to_pe_mesh_dry)
        >> lso_interaction(add_pe_to_pe_mesh_real)
        >> lso_interaction(check_pe_ibgp)
        >> lso_interaction(deploy_routing_instances_dry)
        >> lso_interaction(deploy_routing_instances_real)
        >> lso_interaction(check_l3_services)
        >> lso_interaction(update_sdp_mesh_dry)
        >> lso_interaction(update_sdp_mesh_real)
        >> lso_interaction(add_all_p_to_pe_dry)
        >> lso_interaction(add_all_p_to_pe_real)
        >> lso_interaction(add_pe_to_all_p_dry)
        >> lso_interaction(add_pe_to_all_p_real)
        >> lso_interaction(delete_default_routes_dry)
        >> lso_interaction(delete_default_routes_real)
        >> lso_interaction(remove_isis_overload)
        >> update_subscription_model
        >> resync
        >> done
    )
